#!/usr/bin/env python
# Copyright 2006, CRIM, Revolution Linux Inc.
#
# $Id: initrd.py 2036 2006-08-08 19:32:31Z francis $
#
# Authors : Francis Giraldeau <francis.giraldeau@revolutionlinux.com>
#
# This file is part of the MILLE-XTERM project.
#
#		http://www.revolutionlinux.com/mille-xterm/
#		http://www.mille.ca/
#
# This program is covered by the GNU General Public License.
# See the COPYING file in the top-level MILLE-XTERM directory.
#
# -------------------------------------------------------------------------

"""
Script to build initrd's for booting terminals
"""

from inisettings import *
import sys, os
import tempfile
import time
import popen2
import re
import shutil
import exceptions
from crypt import crypt
import random
from utils import wise_mkdir

success = False
debug = True

class Error(exceptions.EnvironmentError):
	pass

def cryptPassword(password):
	alpha = string.lowercase + string.uppercase
	salt = ""
	for i in range(2):
		index = random.randrange(0, len(alpha))
		salt = salt + alpha[index]
	return crypt(password, salt)

def busybox_utilities(busyboxpath):
	util_list = []
	# Get busybox utilities
	if not os.path.isfile(busyboxpath):
		print "Error : ", busyboxpath, " is not a file "
		return util_list
	stdout, stdin, stderr = popen2.popen3(string.join([busyboxpath,"--help"]))
	lines = stdout.readlines()
	cut = 0
	for num, line in enumerate(lines):
		if line == "Currently defined functions:\n":
			cut = num + 1
	raw_util_list = lines[cut:]
	stdout.close()
	stdin.close()
	stderr.close()
	util_list = [string.strip(x) for x in string.split(string.join(raw_util_list),",")]
	return util_list

def create_devices(dev_dir):
	devices = {"console": "c 5 1",
		   #"fd0": "b 2 0",
		   "null": "c 1 3",
		   "ram0": "b 1 0",
		   "tty0": "c 4 0",
		   "tty1": "c 4 1",
		   "tty2": "c 4 2",
		   "tty3": "c 4 3",
		   "urandom": "c 1 9" }
	for dev in devices.keys():
		dev_path = os.path.join(dev_dir,dev)
		sh(["mknod",dev_path,devices[dev]])

def find_modules_dependencies(modules, modules_path):
	# Find modules dependencies
	# modules = {"mod_name":["path1","path2"]}
	# modules = {"snd-emu10k1":[]}

	file = open(os.path.join(modules_path,"modules.dep"), "r")
	for line in file.readlines():
		# /path/to/modules/module.ko: /path/to/modules/dep.ko
		modline = string.split(line,":")
		mod_str = modline[0]
		for module in modules.keys():
			if module in mod_str:
				modules[module].append(mod_str)
				# Get dependencies
				if len(modline) == 2:
					modules[module] += string.split(modline[1])

def search_file(filename, search_path):
	"""
	Given a search path, find and return the list of file(s) found
	Ex : search_file("e100.ko","/lib/modules/")
	"""
	file_found = False
	paths = string.split(search_path, os.pathsep)
	found_paths = []
	for path in paths:
		for root, dirs, files in os.walk(path):
			for file in files:
				if filename in file:
					file_found = True
					found_paths.append(os.path.join(root,file))
	if file_found:
		return found_paths
	else:
		return None

def copytree2(src,dst,symlinks=0,svn=1):
	"""
	copytree2 is like copytree, but doesn't fail if the destination directory is already present.
	Set the svn argument to 0 to avoid copy the svn cache directories
	"""
	# Do not copy svn directory
	names = []
	if svn : names = os.listdir(src)
	else : names = [x for x in os.listdir(src) if ".svn" not in x]
	#os.mkdir(dst)
	# Prior verifications
	if not os.path.isdir(dst):
		if not os.path.exists(dst):
			os.makedirs(dst)
		else:
			raise IOError
	errors = []
	for name in names:
		srcname = os.path.join(src, name)
		dstname = os.path.join(dst, name)
		try:
			if symlinks and os.path.islink(srcname):
				linkto = os.readlink(srcname)
				os.symlink(linkto, dstname)
			elif os.path.isdir(srcname):
				copytree2(srcname, dstname, symlinks)
			else:
				shutil.copy2(srcname, dstname)
			# XXX What about devices, sockets etc.?
		except (IOError, os.error), why:
			errors.append((srcname, dstname, why))
		# catch the Error from the recursive copytree so that we can
		# continue with other files
		except Error, err:
			errors.extend(err.args[0])
	if errors:
		raise Error, errors

def sh(cmd):
	# FIXME : should call the command with absolute path
	if type(cmd) == list:
		cmd = string.join(cmd)
	if debug:
		print "Executing : ", cmd
	ret = os.system(cmd)
	if ret > 0:
		print "Error while executing the command : ", cmd
	return ret

def modules_ungzip(path): 
	"""Uncompress kernel modules found in a path"""
	count = 0
	for root, dirs, files in os.walk(path):
		for file in files:
			if "o.gz" in file:
				sh(["gzip","-d",os.path.join(root,file)])

	if debug : print "Message : ungzip of " + str(count) + " modules"

class RamDisk:
	def __init__(self, dir, ini_settings):
		self.cur_path = os.path.abspath(os.path.curdir)
		self.ini_settings = ini_settings
		self.settings = ini_settings.settings

		self.dir = dir
		self.mount_dir = "initrd.loop"
		self.mount_dir_path = os.path.join(self.dir,self.mount_dir)
		self.build_dir = "initrd.build"
		self.build_dir_path = os.path.join(self.dir,self.build_dir)
		self.initrd_img = "initrd"
		self.initrd_img_path = os.path.join(self.dir,self.initrd_img)

		# Prior verifications
		wise_mkdir(self.mount_dir_path)
		wise_mkdir(self.build_dir_path)

		# delete previous ramdisk, if any
		self.removeRamdisk()
		
	def getRoot(self):
		return self.build_dir_path
	
	def getInitrdImgPath(self):
		return self.initrd_img_path

	def removeRamdisk(self):
		# be shure to unmount it before removing
		self.umount()
		if os.path.isfile(self.initrd_img_path):
			os.remove(self.initrd_img_path)

	def createRamdisk(self, size, inodes):
		sh(["dd","bs=1k","if=/dev/zero","of="+self.initrd_img_path,"count="+str(size)])
		# Format
		sh(['mke2fs','-F','-q','-m','0','-N',str(inodes), self.initrd_img_path, str(size) ])

	def mount(self):
		ret = os.system(string.join(["mount","-o","loop", \
					      self.initrd_img_path, \
					      self.mount_dir_path]))
		if ret > 0:
			print "Error while mounting the ramdisk :", ret
			sys.exit(1)

	def close(self):
		# Calculate the size of ramdisk to create
		total_size = 0
		total_file = 0

		for r,d,f in os.walk(self.build_dir_path):
			for file in f:
				total_file += 1
				if not os.path.islink(os.path.join(r,file)):
					total_size += os.stat(os.path.join(r,file))[6]

		size = (total_size / 1000) + 500
		inodes = total_file + 500

		# Create the ramdisk and mount it
		self.createRamdisk(size,inodes)
		self.mount()
		
		# Move things into ramdisk
		sh(['cp -af',self.build_dir_path+"/*",self.mount_dir_path])

		self.umount()
		
	def umount(self):
		# May be mounted multiple times
		count = 0
		while os.path.ismount(self.mount_dir_path) and count < 10:
			ret = sh(["umount",self.mount_dir_path])
			count += 1
			if ret > 0:
				print "Error while unmounting the ramdisk"
				time.sleep(0.5)
	
	def install(self,src,dst):
		if len(dst) > 0:
			if dst[0] == "/":
				dst = dst[1:]
		abs_dst = os.path.join(self.build_dir_path,dst)
		if not os.path.isdir(abs_dst):
			print "Warning : not a directory : ", abs_dst
			if not os.path.exists(abs_dst):
				print "Creating the directory"
				os.makedirs(abs_dst)
			else:
				"Error : can't create directory : ", abs_dst
				sys.exit(1)
				
		if os.path.isfile(src):
			if debug: print "Copy file : ", src, " to ", abs_dst
			shutil.copy(src, abs_dst)
		elif os.path.isdir(src):
			if debug : print "Copy tree : ", src, " to ", abs_dst
			copytree2(src,abs_dst,svn=0)
		else:
			"Error : install file or directory only. ", src

	def link(self,target,link):
		"""
		Create relative links that points to target into the ramdisk
		Ex : link("/bin/busybox","/usr/bin/ping") will do : ln -s ../../bin/busybox usr/bin/ping
		"""
		os.chdir(self.build_dir_path)
		
		# Remove / from the link if any
		if link[0] == os.path.sep:
			link = link[1:]

		# Remove / from the link if any
		if target[0] == os.path.sep:
			target = target[1:]

		# Test if the target exists
		if not os.path.isfile(os.path.join(".",target)):
			print "Error : not a file : ", os.path.join(self.build_dir_path,target)
			os.chdir(self.cur_path)
			return
			
		# Create the link directory if it doesn't exists
		link_dir, link_base = os.path.split(link)
		if not os.path.isdir(link_dir):
			print "Warning : not a directory : ", link_dir
			if not os.path.exists(link_dir):
				print "Creating the directory"
				os.makedirs(link_dir)
			else:
				"Error : can't create directory : ", link_dir
		
		# Get the deep of the link, and add .. to the target
		dir, basename = os.path.split(link)
		
		for i in range(len(string.split(dir,os.path.sep))):
			target = os.path.join(os.pardir,target)

		# Do the link
		try : 
			if debug: print "Linking : target=", target, " link=", link 
			os.symlink(target,link)
		except :
			print "Error : ", str(sys.exc_value)

		os.chdir(self.cur_path)

			
	def __del__(self):
		if self.mount_dir:
			os.chdir(self.cur_path)
			self.umount()

class Initrd:
	def __init__(self, ini_settings):
		"""Initrd class"""
		self.ini_settings = ini_settings
		self.settings = ini_settings.settings

		# kernel related paths
		# FIXME : xtroot should be a parameter
		self.xtroot="/opt/xtermroot"
		self.kernel_image_path = os.path.join(self.xtroot,"boot")
		self.kernel_modules_path = os.path.join(self.xtroot,"lib/modules")
		
		# Create temporary working directory and timestamp
		tmp_path = tempfile.mkdtemp()
		stdout, stdin = popen2.popen2("date +%Y%m%d%H%M%S")
		self.timestamp = string.strip(stdout.readline())
		stdin.close()
		stdout.close()
		self.initrd_dir = os.path.join(tmp_path, self.timestamp) 
		os.mkdir(self.initrd_dir)
		print "Initrd directory will be : ", self.initrd_dir

		# Set default values
		self.ini_settings.setDefault("initrd","kernel", self.findKernel())
		
		# Validate the kernel, if the kernel is not found
		# Try to find another one
		if not self.validateKernel():
			print "Warning : specified kernel not found."
			self.settings["initrd"]["kernel"] = self.findKernel()
			print "Warning : Using kernel " + self.settings["initrd"]["kernel"] + " instead"

		self.ini_settings.save()

		self.ramdisk = RamDisk(self.initrd_dir,self.ini_settings)
		self.ramdisk_path = self.ramdisk.getRoot()

		# Set the path of initroot
		# Let the devel directory have precedence
		self.initrd_root = ""
		if  os.path.isdir("data"):
			self.initrd_root = "data"
		elif os.path.isdir("/var/lib/mille-xterm-bootserver"):
			self.initrd_root = "/var/lib/mille-xterm-bootserver"
		
		self.initrd_root = os.path.join(self.initrd_root,"initroot")

	def getInitrdImgPath(self):
		return self.ramdisk.getInitrdImgPath()

	def getInitrdImgName(self):
		return os.path.basename(self.ramdisk.getInitrdImgPath())

	def copyBinary(self, binary):
		# Copy all library needed to run a binary

		cmd = string.join(["ldd",binary])

		stdout, stdin = popen2.popen2(cmd)

		libs_raw = [ string.strip(x) for x in stdout.readlines()[1:-1] ]

		# Get libs target and links
		# libs = {target:link}
		libs = {}
		for lib in libs_raw:
			lib_line = string.split(lib)
			libs[lib_line[2]] = lib_line[0]

		# install libs and make links
		for lib in libs.keys():
			d, f = os.path.split(lib)
			self.ramdisk.install(lib, d)
			self.ramdisk.link(lib, os.path.join("lib",libs[lib]) )
				
		d, f = os.path.split(binary)
		self.ramdisk.install(binary,d)
		cmd = string.join(["ldconfig","-r",self.ramdisk.getRoot()])
		stdout, stdin = popen2.popen2(cmd)

	def findKernel(self):
		"""Find suitable kernel installed on the system"""
		kernels = os.listdir(self.kernel_modules_path)
		if len(kernels) == 0:
			print "Error : no kernels in " + self.kernel_modules_path
			sys.exit(1)
		# Remove not suitable kernels
		good_kernels = []
		for k in kernels:
			if "686" not in k and "smp" not in k and "enterprise" not in k:
				# We must have kernel 2.6.14 for fuse to work
				# out of the box
				if k >= "2.6.14":
					good_kernels.append(k)
		# By default, get the latest kernel
		if len(good_kernels) == 0:
			print "Error : no suitable kernels in " + self.kernel_modules_path
			sys.exit(1)

		good_kernels.sort()
		return good_kernels[-1] 

	def validateKernel(self):
		"""Validate the choosen kernel"""
		self.kernel_version = self.settings["initrd"]["kernel"]
		self.kernel_image = os.path.join(self.kernel_image_path,"vmlinuz-" + self.kernel_version)
		self.kernel_map = os.path.join(self.kernel_image_path,"System.map-" + self.kernel_version)
		self.kernel_modules = os.path.join(self.kernel_modules_path,self.kernel_version)
		# Barely verify that the kernel and modules exists
		#print self.kernel_image
		#print self.kernel_modules
		if os.path.isfile(self.kernel_image) and \
		   os.path.isfile(self.kernel_map) and \
		   os.path.isdir(self.kernel_modules):
			# Kernel valid
			return True
		else :
			return False

	def findModulePath(self, mod):
		module = mod + ".ko"
		search_path = os.path.join(self.kernel_modules_path,self.kernel_version)
		path_list = search_file(module, search_path )
		if path_list:
			return path_list
		else:
			print "Warning : Module ", module, " not found in initrd, skipping" 
			return path_list
			
	def buildInitrd(self):
		"""Function that performs all operations for building the initrd"""
		step_names = ["InstallStatic","InstallDynamic","Busybox","Modules","Devices","Configure","Finalize"]
		for step_name in step_names:
			print "#### Starting step : ", step_name , " #### "
			eval("self.step" + step_name)()
		return True

	def stepInstallStatic(self):
		"""Copy needed files to the initrd"""
		# List of binaries to copy
		# FIXME : should be created dynamicaly
		# We should copy the binary, not only a link
		files = ["/sbin/pci_scan"]

		for file in files:
			d, f = os.path.split(file)
			self.ramdisk.install(file,d)
		
		# FIXME: where the files should be ?
		# Temporary 
		self.ramdisk.install(self.initrd_root,"")
	
	def stepInstallDynamic(self):
		"""Copy binary files and their dependancies to the initrd"""
		return

	def stepBusybox(self):
		"""Copy and symlink busybox in the initrd"""
		# Locate the busybox binary
		self.busybox_path = "/bin/busybox"
		if not os.path.isfile(self.busybox_path):
			print "Busybox not found in ", self.busybox_path
			sys.exit(1)

		util_list = busybox_utilities(self.busybox_path)
		# Copy busybox
		self.ramdisk.install(self.busybox_path,"/bin")
		# For each util, do symlink in the initrd
		# Everything is in /bin, because there is no way to know in wich directory
		# to link the utility
		for util in util_list:
			self.ramdisk.link(self.busybox_path,os.path.join("/bin",util) )
			self.ramdisk.link(self.busybox_path,os.path.join("/sbin",util) )
			
	def stepModules(self):
		"""Copy kernel modules and perform related operations"""
		# FIXME : Should use a more general mechanism than to use the niclist

		# Build the module list to copy
		# On each line, there is a pci id, the network driver and a comment
		# We append only if the driver is not in the modlist already
		# FIXME : global variables should be keep into __init__.py file
		name = os.path.join(self.initrd_root,"etc/niclist")
		file = open(name,"r")
		# Drop comments
		lines = [ string.strip(x) for x in file.readlines() if x[0] is not "#" ]

		# modlist { module_name : module_path }
		modules = {"af_packet":[],
			   "sunrpc":[],
			   "nfs":[],
			   "lockd":[]
		}
		for line in lines:
			s = string.split(line)
			# Take care of malformed lines
			if len(s) > 2:
				# append if no such module
				if s[1] not in modules.keys():
					modules[s[1]] = []
					
		find_modules_dependencies(modules, self.kernel_modules)
		
		# Merge lists of modules
		merged_module_list = []
		for module in modules.values():
			for m in module:
				if m not in merged_module_list:
					merged_module_list.append(m)
		
		# Copy modules into the initrd
		if debug : print merged_module_list
		for module in merged_module_list:
			# remove "/"
			if len(module) > 0:
				if module[0] == os.path.sep:
					module = module[1:]
			self.ramdisk.install(os.path.join(self.xtroot,module), os.path.join("lib/modules",self.kernel_version) )


		# make shure that modules are ungzip
		mp = os.path.join(self.ramdisk.getRoot(),"lib/modules")
		modules_ungzip(mp)
	
		# Create modules.dep file
		# JMD: we need depmod-25 for use in vserver
		if os.path.exists("/sbin/depmod-25"):
		  sh(string.join(["/sbin/depmod-25","-a","-e","-F",self.kernel_map,"-b",self.ramdisk.getRoot(),self.kernel_version]))
		else:
		  sh(string.join(["/sbin/depmod","-a","-e","-F",self.kernel_map,"-b",self.ramdisk.getRoot(),self.kernel_version]))
		
	def stepDevices(self):
		dev_dir = os.path.join(self.ramdisk.getRoot(),"dev")
		if not os.path.isdir(dev_dir):
			print "Warning : not a directory : ", dev_dir
			if not os.path.exists(dev_dir):
				print "Creating the directory"
				os.makedirs(dev_dir)
			else:
				"Error : can't create directory : ", dev_dir

		create_devices(dev_dir)
		
	def stepConfigure(self):
		# Setup root environment
		# FIXME : Should be in the mille-xterm configuration
		root_password = "coucou"
		crypt_root_password = cryptPassword(root_password)

		file = open(os.path.join(self.ramdisk.getRoot(),"etc","passwd"),"w")
		file.write("root:x:0:0:root:/:/bin/ash\n")
		file.close()

		file = open(os.path.join(self.ramdisk.getRoot(),"etc","shadow"),"w")
		file.write("root:"+crypt_root_password+":12579:0:99999:7:::\n")
		file.close()

		file = open(os.path.join(self.ramdisk.getRoot(),"etc","group"),"w")
		file.write("root:x:0:\n")
		file.close()

	def stepFinalize(self):
		self.ramdisk.close()

		# Now, copy the kernel and the ramdisk into tftpboot
		# Copy in every tftpboot dirs available
		tftpboot_dirs = ["/tftpboot","/var/lib/tftpboot/X86PC/linux/"]
		for tftpboot_dir in tftpboot_dirs:
			if os.path.isdir(tftpboot_dir):
				shutil.copy2(self.getInitrdImgPath(),tftpboot_dir)
				sh(["gzip","-f",os.path.join(tftpboot_dir, self.getInitrdImgName())])
				shutil.copy2(self.kernel_image,os.path.join(tftpboot_dir,"vmlinuz"))
				
		# Setup the update directory
		updates_dir = os.path.join("/opt","xtermroot","updates",self.timestamp)
		print "Setting up the update directory in ", updates_dir
		print "Don't forget to synchronize updates directory over the cluster."
		# Copy initrd and kernel in update directory
		sh(["mkdir","-p",updates_dir])
		shutil.copy2(self.kernel_image, os.path.join(updates_dir,"vmlinuz"))
		shutil.copy2(self.getInitrdImgPath(), updates_dir)
		pushd = os.path.abspath(os.path.curdir)
		
		# Setup the link to current version
		os.chdir(os.path.join("/opt","xtermroot","updates"))
		if os.path.islink("current"):
			os.remove("current")
		os.symlink(self.timestamp,"current")

		# Put the version into version file
		print "writing version file for updates"
		os.chdir(updates_dir)
		file = open(os.path.join(updates_dir,"version"),"w")
		file.write(self.timestamp)
		file.close
		os.chdir(pushd)

		print "Build initrd is over"
		

import unittest 
class TestInitrd(unittest.TestCase):
	def setUp(self):
		self.current_kernel = os.uname()[2]
		self.ini_settings = IniSettings('../bootserver.ini')
		self.initrd = Initrd(self.ini_settings)
	def tearDown(self):
		del self.initrd
	def testFindKernel(self):
		assert "2.6" in self.initrd.findKernel()
	def testValidateKernel1(self):
		self.assert_(self.initrd.validateKernel())
	def testValidateKernel2(self):
		self.initrd.settings["initrd"]["kernel"] = "kernel-1.2.3"
		assert self.initrd.validateKernel() == False
	def testBootserverSettings1(self):
		assert self.ini_settings.settings.has_key("kernel") == False
		assert self.initrd.validateKernel() == True
		self.ini_settings.save()
		self.ini_settings = IniSettings('../bootserver.ini')
		assert self.ini_settings.settings["initrd"].has_key("kernel") == True
	def testStepBusybox(self):
		self.initrd.stepBusybox()
	def testRamDisk(self):
		tmp_dir = "/tmp/ramdisk-test"
		ramdisk = RamDisk(tmp_dir,self.ini_settings)
	def testRamDisk2(self):
		tmp_dir = "/tmp/ramdisk-test"
		ramdisk = RamDisk(tmp_dir,self.ini_settings)
		ramdisk.install("/bin/ls","bin")
		ramdisk.link("/bin/ls","/bin/ls2")
	def testStepModules(self):
		self.initrd.stepModules()
	def testStepDevices(self):
		self.initrd.stepDevices()
	def testCopyBinary(self):
		self.initrd.copyBinary("/sbin/udevd")
		self.initrd.copyBinary("/sbin/udevstart")

class TestFastInitrd(unittest.TestCase):
	def setUp(self):
		self.current_kernel = os.uname()[2]
		self.ini_settings = IniSettings('../bootserver.ini')
		self.initrd = Initrd(self.ini_settings)

if __name__=="__main__":

	# This program must be run as root
	if os.getuid() != 0:
		print "This program must me run as root"
		sys.exit(1)
	
	if len(sys.argv)>1:
		if sys.argv[1] == "unittest":
			suite = unittest.makeSuite(TestInitrd)
			#suite = unittest.makeSuite(TestFastInitrd)
			unittest.TextTestRunner(verbosity=2).run(suite)
			sys.exit(0)

	ini_settings = IniSettings('../bootserver.ini')
	initrd = Initrd(ini_settings)
	initrd.buildInitrd()

"""

  echo "Installing RPMS..."
  for i in busybox*	\
	 insmod-static* \
	 module-init-tools* \
	 pci_scan*	\
	 slptool-static* 
  do
   $SUDO  rpm --root ${DESTDIR} --excludedocs -ivh $RPMS/${i}
  done

  # JMD: no links for insmod, lsmod, modprobe and rmmod since we provide
  # static binaries that have gzip support

  echo "Installing module tools"
  pushd ${DESTDIR}/sbin >/dev/null
  for link in insmod.old lsmod.old modprobe.old rmmod.old;do ln -sf insmod.static $link;done
  for link in insmod lsmod modprobe rmmod;do ln -sf ${link}2.6-static $link;done
  popd >/dev/null

  
  # Setup the device nodes
  #
  echo "Setup the /dev directory..."

  [ -e ${DESTDIR}/dev/console ] || $SUDO mknod ${DESTDIR}/dev/console  c 5 1
  [ -e ${DESTDIR}/dev/fd0     ] || $SUDO mknod ${DESTDIR}/dev/fd0      b 2 0
  [ -e ${DESTDIR}/dev/null    ] || $SUDO mknod ${DESTDIR}/dev/null     c 1 3
  [ -e ${DESTDIR}/dev/ram0    ] || $SUDO mknod ${DESTDIR}/dev/ram0     b 1 0
  [ -e ${DESTDIR}/dev/tty0    ] || $SUDO mknod ${DESTDIR}/dev/tty0     c 4 0
  [ -e ${DESTDIR}/dev/tty1    ] || $SUDO mknod ${DESTDIR}/dev/tty1     c 4 1
  [ -e ${DESTDIR}/dev/tty2    ] || $SUDO mknod ${DESTDIR}/dev/tty2     c 4 2
  [ -e ${DESTDIR}/dev/tty3    ] || $SUDO mknod ${DESTDIR}/dev/tty3     c 4 3

  #
  # Setup root environment
  #
  echo "root:x:0:0:root:/:/bin/ash" > ${DESTDIR}/etc/passwd
  echo "root:\$1\$\$qRPK7m23GJusamGpoGLby/:12579:0:99999:7:::" > ${DESTDIR}/etc/shadow
  echo "root:x:0:" > ${DESTDIR}/etc/group

  #
  # Put debug info into the initrd
  #
  
  sed -e "s/DEBUG_LEVEL/${DEBUG_LEVEL}/" ${DESTDIR}/linuxrc > ${DESTDIR}/linuxrc.tmp
  mv -f  ${DESTDIR}/linuxrc.tmp  ${DESTDIR}/linuxrc
  chmod +x ${DESTDIR}/linuxrc

  pushd  ${MODULE_DIR} >/dev/null
  if [ ! -d ${DESTDIR}${MODULE_DIR} ]; then
      mkdir -p ${DESTDIR}${MODULE_DIR}
  fi
  
  
  if [ "${CONF_BUSTYPE}" = "pci_isa" ]; then

    #Create module directory
    mkdir -p ${DESTDIR}${MODULE_DIR}/kernel/drivers/net
    mkdir -p ${DESTDIR}${MODULE_DIR}/kernel/3rdparty
    mkdir -p ${DESTDIR}${MODULE_DIR}/kernel/net/packet
    
    # af_packet is required by some network cards?
    cp -af kernel/net/packet/* ${DESTDIR}${MODULE_DIR}/kernel/net/packet
    
    # copy only PCI drivers
    # JMD: dgrs is a multiport card, acenic is gigabit, skfp is FDDI, don't
    # include them
    for driver in \
     3c59x 8139too 8390 defxx dl2k dmfe e100 e1000 \
     eepro100 epic100 fealnx forcedeth hamachi hp100 mii natsemi ne2k-pci ns83820 \
     pcnet32 r8169 rcpci sis900 starfire sundance sungem sunhme tg3 tlan \
     tulip typhoon via-rhine winbond-840 yellowfin 
    do 
     cp -af kernel/drivers/net/$driver* ${DESTDIR}${MODULE_DIR}/kernel/drivers/net
    done
  
    cp -af kernel/3rdparty/bcm* ${DESTDIR}${MODULE_DIR}/kernel/3rdparty

    #JMD: needed for 2.6 kernels
    mkdir -p ${DESTDIR}${MODULE_DIR}/kernel/lib
    cp -af kernel/lib/* ${DESTDIR}${MODULE_DIR}/kernel/lib
    

  fi
      
  if [ "${CONF_BUSTYPE}" = "pcmcia" ]; then
      #
      # Setup the directory for holding the kernel pcmcia modules
      #
      echo "mkdir -p /mnt/initrd${MODULE_DIR}/pcmcia"
      mkdir -p ${DESTDIR}${MODULE_DIR}/pcmcia
      if [ $? -ne 0 ]; then
          write_error "Creating pcmcia module directory"
      fi

      for i in   3c574_cs.o*      \
                 3c575_cb.o*      \
                 3c589_cs.o*      \
                 airo_cs.o*       \
                 airo.o*          \
                 axnet_cs.o*      \
                 cb_enabler.o*    \
                 ds.o*            \
                 dummy_cs.o*      \
                 eepro100_cb.o*   \
                 epic_cb.o*       \
                 hermes.o*        \
                 i82365.o*        \
                 netwave_cs.o*    \
                 nmclan_cs.o*     \
                 orinoco_cs.o*    \
                 orinoco.o*       \
                 pcmcia_core.o*   \
                 pcnet_cs.o*      \
                 ray_cs.o*        \
                 smc91c92_cs.o*   \
                 tcic.o*          \
                 tulip_cb.o*      \
                 wavelan_cs.o*    \
                 wvlan_cs.o*      \
                 xirc2ps_cs.o*
      do
          cp pcmcia/${i} ${DESTDIR}${MODULE_DIR}/pcmcia/
          if [ $? -ne 0 ]; then
              write_error "Copying $i"
          fi
      done


      #
      # Grab the 8390 module, it's needed by pcnet_cs
      #
      find ./kernel/drivers/net -name 8390.o* -print | \
          cpio -pmud ${DESTDIR}${MODULE_DIR}/
          if [ $? -ne 0 ]; then
              write_error "Copying 8390.o*"
          fi

  fi

  #copy filesystem drivers
  mkdir -p ${DESTDIR}${MODULE_DIR}/kernel/fs
  cp -af kernel/fs/*fat* ${DESTDIR}${MODULE_DIR}/kernel/fs
  cp -af kernel/fs/isofs* ${DESTDIR}${MODULE_DIR}/kernel/fs
  cp -af kernel/fs/msdos* ${DESTDIR}${MODULE_DIR}/kernel/fs
  mkdir -p ${DESTDIR}${MODULE_DIR}/kernel/fs/nls
  cp -af kernel/fs/nls/nls_cp437* ${DESTDIR}${MODULE_DIR}/kernel/fs/nls
  cp -af kernel/fs/nls/nls_iso8859-1.* ${DESTDIR}${MODULE_DIR}/kernel/fs/nls
    
  #copy block drivers
  mkdir -p ${DESTDIR}${MODULE_DIR}/kernel/drivers/block
  cp -af kernel/drivers/block/floppy* ${DESTDIR}${MODULE_DIR}/kernel/drivers/block
  mkdir -p ${DESTDIR}${MODULE_DIR}/kernel/drivers/cdrom
  cp -af kernel/drivers/cdrom/cdrom* ${DESTDIR}${MODULE_DIR}/kernel/drivers/cdrom
  mkdir -p ${DESTDIR}${MODULE_DIR}/kernel/drivers/ide
  cp -af kernel/drivers/ide/ide-cd* ${DESTDIR}${MODULE_DIR}/kernel/drivers/ide
  cp -af kernel/drivers/ide/ide-floppy* ${DESTDIR}${MODULE_DIR}/kernel/drivers/ide
  
  #copy agpgart driver
  mkdir -p ${DESTDIR}${MODULE_DIR}/kernel/drivers/char
  cp -af kernel/drivers/char/agp ${DESTDIR}${MODULE_DIR}/kernel/drivers/char

  #copy iptables
  #mkdir -p ${DESTDIR}${MODULE_DIR}/kernel/net/ipv4/netfilter
  #cp -af kernel/net/ipv4/netfilter/* ${DESTDIR}${MODULE_DIR}/kernel/net/ipv4/netfilter
           
  popd >/dev/null

  # 
  # Gunzip the modules and give them the right permissions
  #
  find ${DESTDIR}${MODULE_DIR} -type f|xargs -n1 gunzip
  find ${DESTDIR} -type f|xargs $SUDO chown root.root

  #
  # build the modules.dep file so that modprobe will work
  #
  depmod -e -a -F ${SYSTEM_MAP} -b /mnt/initrd ${KERNEL_VER}
  if [ $? -ne 0 ]; then
      write_error "depmod"
  fi


  echo 
  echo "Checking size"
  df -v /mnt/initrd
  echo 
  df -i /mnt/initrd
  echo 

  #
  # Un-mount the filesystem
  #
  $SUDO umount /mnt/initrd

  #
  # Compress it, to prepare it to be tacked onto the
  # end of the tagged kernel
  #
  echo "gzipping initrd"
  if [ "${CONF_BUSTYPE}" = "pcmcia" ]; then
      INITRD_OUTPUT=${CURDIR}/initrd-pcmcia-${KERNEL_VER}.gz
  else
      INITRD_OUTPUT=${CURDIR}/initrd-${KERNEL_VER}.gz
  fi

  gzip -9 <${CURDIR}/${INITRD_IMG} >${INITRD_OUTPUT}

  if [ -x /usr/share/bootsplash/scripts/make-boot-splash ];then
     /usr/share/bootsplash/scripts/make-boot-splash ${INITRD_OUTPUT} 800x600 2>/dev/null
  fi

  cp ${KERNEL_IMG} ${CURDIR}/bzImage-${KERNEL_VER}

  if [ "${CONF_BUSTYPE}" = "pcmcia" ]; then
      exit
  fi

  echo "running mknbi-linux"
  if [ "${CONFIG_LPP}" = "Y" ]; then
      ${MKNBI} --output=${CURDIR}/${KERNEL_OUTPUT}                     \
      		--append="init=/linuxrc rw"				\
	       --rootdir="/dev/ram0"					\
               ${KERNEL_IMG}                                           \
               ${INITRD_OUTPUT}
  else
      ${MKNBI} --output=${CURDIR}/${KERNEL_OUTPUT}                     \
	       --append="root=/dev/ram0 init=/linuxrc panic=15 rw vga=788 media=etherboot menu=terminal INITRD_DBG=${DEBUG_LEVEL}" \
               ${KERNEL_IMG}                                           \
               ${INITRD_OUTPUT}
  fi

  # Show the size of the kernel+initrd image
  #
  ls -lsh ${CURDIR}/${KERNEL_OUTPUT}

  #
  # Move the stuff into tftpboot
  #
  
  if [ ! -d $IMAGES/${KERNEL_VER} ]; then
	  mkdir -p $IMAGES/${KERNEL_VER}
  fi
  mv ${CURDIR}/bzImage-${KERNEL_VER} $IMAGES/${KERNEL_VER}/vmlinuz
  mv ${INITRD_OUTPUT} $IMAGES/${KERNEL_VER}/initrd
  mv ${CURDIR}/${KERNEL_OUTPUT} $IMAGES/${KERNEL_VER}/boot.nbi
  echo $DAT > $IMAGES/${KERNEL_VER}/version

  rm -rf ./tmp/iso
  mkdir -p ./tmp/iso/boot/default
  mkdir -p ./tmp/iso/boot/grub
  cp $IMAGES/${KERNEL_VER}/vmlinuz ./tmp/iso/boot/default
  cp $IMAGES/${KERNEL_VER}/initrd ./tmp/iso/boot/default
  cp -f ./grub/* ./tmp/iso/boot/grub
  
  # Setting up the grub menu
  cat ./cdrom-menu.lst > ./tmp/iso/boot/grub/menu.lst
  
    
  # Taken verbatim from:
  # http://www.gnu.org/software/grub/manual/html_node/Making-a-GRUB-bootable-CD-ROM.html
  mkisofs -R -b boot/grub/stage2_eltorito -no-emul-boot \
	-boot-load-size 4 -boot-info-table -o $IMAGES/${KERNEL_VER}/grub.iso ./tmp/iso

  cat ./cdrom-menu-1067.lst > ./tmp/iso/boot/grub/menu.lst

  mkisofs -R -b boot/grub/stage2_eltorito -no-emul-boot \
	-boot-load-size 4 -boot-info-table -o $IMAGES/${KERNEL_VER}/grub-1067.iso ./tmp/iso
  
rm -rf ./tmp/iso
rm -f initrd.mille config.in
    
echo

echo "Copying kernel into $LTSPDIR"
$SUDO mkdir -p ${LTSPDIR}/lib/modules
$SUDO cp -a /lib/modules/$KERNEL ${LTSPDIR}/lib/modules
$SUDO touch -r ${LTSPDIR}/lib/modules/$KERNEL/modules.dep ${LTSPDIR}/etc/modules.conf

echo
echo 
echo "Setting up the update directory in ${LTSPDIR}/updates"
echo "Don't forget to synchronize updates directory over the cluster."
echo 
$SUDO mkdir -p ${LTSPDIR}/updates/$DAT/
$SUDO cp -f $IMAGES/${KERNEL_VER}/vmlinuz ${LTSPDIR}/updates/$DAT
$SUDO cp -f $IMAGES/${KERNEL_VER}/initrd  ${LTSPDIR}/updates/$DAT
pushd ${LTSPDIR}/ >/dev/null
$SUDO rm -f ./updates/current
$SUDO ln -sf ../updates/$DAT ./updates/current
popd >/dev/null
$SUDO sh -c "echo $DAT > ${LTSPDIR}/updates/$DAT/version"

eval export `cat ../../VERSION | grep ^VERSION=`


echo 
echo
mille_log "Building initrd done. You will find images in $IMAGES/${KERNEL_VER}"
echo
echo
"""
