#!/bin/sh
# 
# Copyright 2006, CRIM, Revolution Linux Inc.
#
# $Id$
#
# Authors : Ivan Arsenault Ivan.Arsenault@crim.ca
#
# This file is part of the MILLE-XTERM project.
#
#       http://www.revolutionlinux.com/mille-xterm/
#       http://www.mille.ca/
#
# This program is covered by the GNU General Public License.
# See the COPYING file in the top-level MILLE-XTERM directory.
#
# -------------------------------------------------------------------------
#
#Xterm-shaper is a script that creates traffic managing rules for X terminals.
#It is to be used within the mille-xterm or ltsp projects, to put bandwidth
#limits to different network protocols.
#
#Functions and parameters
#¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯
#A) Show status
#	param1 : the "status" command
#	param2 : the network interface
#	EX: # xterm-shaper.sh status eth0
#B) Clear
#	param1 : the "clear" command
#	param2 : the network interface
#	EX: # xterm-shaper.sh clear eth0
#C) Create a traffic managing rule
#	param1 : the network interface
#	param2 : maximum speed to be allowed to any other protocol than those defined in param4
#	param3 : default priority
#param4 & more : traffic control definitions:
# 		 list of parameters between quotes, where each
# 		 parameter is seperated by a "|".
# 		 for a range of ports, use ":" to seperate the two port numbers.
#		 ex.1: "port|speed|maximum_speed|priority"
#		 ex.2: "port1:to_port2|spped|maximum_speed|priority|unique_sequential_number"
#  Note:  - The speed should be set in terms of kilobits per second.
#         - This script should be used after having set up firewall
#           rules, i.e.: we suppose that firewall policies have
#           already been applied, or that they are active by default.
#												        
#Creating a traffic managing rule:
#¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯
#EXAMPLE: term-shaper.sh eth0 100000 3 "22|10|15|1|101" "6000:6009|10000|15000|2|102"
# 	   Criterion1: ssh (port 22) is set at priority 1 and uses between 
#                      10kbits and 15kbits of bandwidth.
#  	   Criterion2: X11 (ports 6000 to 6009) is set at priority 2 and uses
#                      betweem 10mbits and 15mbits of bandwidth.
#          Criterion3: By default, all other traffic has no bandwidth limit,
#    		       but we can set it's priority to "3".
#
#Possible improvements				
#¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯¯
#     	- Adding more parameter verification.
##


# Commands that are used in the script
FWCMD="iptables"
TCCMD="tc"
IFCFG="ifconfig"

# Give information on traffic control rules (show status)
#
# Param1 : the network interface
#
function getStatus() {

	netDevice=$1

	echo "*** [qdisc] ***"
	$TCCMD -s qdisc ls dev $netDevice
	echo
	echo "*** [class] ***"
	$TCCMD -s class ls dev $netDevice
	echo "*** [iptable chain XTERM-MANGLE] ***"
	$FWCMD -t mangle -L XTERM-MANGLE -v -x 2> /dev/null
	echo
}

# Clean up everything that concerns traffic control (clear)
#
# Param1 : the network interface
#
function cleanUp() {

	netDevice=$1

	# Clear all traffic control rules
	$TCCMD qdisc del dev $netDevice root    2> /dev/null > /dev/null
	$TCCMD qdisc del dev $netDevice ingress 2> /dev/null > /dev/null

	# Clear all firewall markings in the XTERM-MANGLE chain
	$FWCMD -t mangle -D POSTROUTING -o $netDevice -j XTERM-MANGLE 2> /dev/null > /dev/null
	$FWCMD -t mangle -F XTERM-MANGLE 2> /dev/null > /dev/null
	$FWCMD -t mangle -X XTERM-MANGLE 2> /dev/null > /dev/null
}

# Verify status of traffic control rules (show status)
if [ "$1" = "status" ]; then

	getStatus $2
	exit
fi

# Clear all and exit
if [ "$1" = "clear" ]; then

	cleanUp $2
	exit
fi

# Verify parameters 
if [ -z "$1" ]; then

	echo "Parameter 1 is missing, (network interface)"
	exit 1
fi

if [ -z "$2" ]; then

	echo "Parameter 2 is missing, (maximum speed in kbit)"
	exit 1
fi

if [ -z "$3" ]; then

	echo "Parameter 3 is missing, (priority)"
	exit 1
fi

if [ -z "$4" ]; then

	echo "Parameter 4 is missing, (\"port|speed|maximum_speed|priority|unique_sequential_number\")"
	exit 1
fi

DEV=$1
MAX_SPEED=$2
DEFAUT_PRIO=$3

# Clean up all traffic control rules
cleanUp $DEV

# Create the XTERM-MANGLE chain for the marking of packets
$FWCMD -t mangle -N XTERM-MANGLE
# Associate the XTERM-MANGLE chain to POSTROUTING
$FWCMD -t mangle -I POSTROUTING -o $DEV -j XTERM-MANGLE

# Set queue length for $DEV
$IFCFG $DEV txqueuelen 16

# Create queue on device "$DEV" using the HTB classification method, with "1" as the default class.
$TCCMD qdisc add dev $DEV parent root handle 1:0 htb default 1 r2q 1

# Attach class "1:1" to (previously created) queue "1:0" (of HTB type)
# and set the main limit to "$DEV_MAX_SPEED" with no possible burst.
$TCCMD class add dev $DEV parent 1:0 classid 1:1 htb rate ${MAX_SPEED}kbit \
       ceil ${MAX_SPEED}kbit burst 0 cburst 0 prio $DEFAUT_PRIO

# Until there are control rules...
until [ -z "$4" ]
do

	PORT=`echo $4 | cut -f1 -d "|"`
	SPEED=`echo $4 | cut -f2 -d "|"`
	MAX_SPEED=`echo $4 | cut -f3 -d "|"`
	PRIO=`echo $4 | cut -f4 -d "|"`
	BURST=$(expr $MAX_SPEED - $SPEED)
	MARK=`echo $4 | cut -f5 -d "|"`

	# Mark the packets
	$FWCMD -A XTERM-MANGLE -t mangle -p tcp --dport $PORT \
	       -j MARK --set-mark $MARK 
	$FWCMD -A XTERM-MANGLE -t mangle -p tcp --sport $PORT \
	       -j MARK --set-mark $MARK
	$FWCMD -A XTERM-MANGLE -t mangle -p udp --dport $PORT \
	       -j MARK --set-mark $MARK 
	$FWCMD -A XTERM-MANGLE -t mangle -p udp --sport $PORT \
	       -j MARK --set-mark $MARK

 	# Define the class
	$TCCMD class add dev $DEV parent 1:1 classid 1:$MARK htb \
	         rate ${SPEED}kbit ceil ${MAX_SPEED}kbit burst ${BURST}k \
			 cburst 0b prio $PRIO

	# Filtering of packets
	# Associate the markings done by iptables to the correct class
	$TCCMD filter add dev $DEV parent 1:0 protocol ip pref 0 \
	       handle $MARK fw flowid 1:$MARK

	# Set a rule on queue (of SFQ type) for an equal treatment between classes,
	# i.e.: to make sure that a high-priority class does not use all of the bandwidth
	$TCCMD qdisc add dev $DEV parent 1:$MARK sfq perturb 16

	shift
done
